#include "core/Box.h"

#include "core/Generic.h"

void setBox(Box* box, const Vector3* size) {
    for(size_t i = 0; i < 3; i++) {
        if(size->data[i] < 0.0f) {
            box->min.data[i] = size->data[i];
            box->max.data[i] = 0.0f;
        } else {
            box->min.data[i] = 0.0f;
            box->max.data[i] = size->data[i];
        }
    }
}

void offsetBox(Box* box, const Vector3* offset) {
    addSet(&box->min, offset);
    addSet(&box->max, offset);
}

bool collidesWithBox(const Box* box, const Box* other) {
    return box->max.data[0] > other->min.data[0] &&
           box->min.data[0] < other->max.data[0] &&
           box->max.data[1] > other->min.data[1] &&
           box->min.data[1] < other->max.data[1] &&
           box->max.data[2] > other->min.data[2] &&
           box->min.data[2] < other->max.data[2];
}

void expandBox(Box* box, const Vector3* offset) {
    for(size_t i = 0; i < 3; i++) {
        if(offset->data[i] > 0.0f) {
            box->max.data[i] += offset->data[i];
        } else {
            box->min.data[i] += offset->data[i];
        }
    }
}

void growBox(Box* box, const Vector3* growth) {
    Vector3 half;
    mul(&half, growth, 0.5f);
    Vector3 nMin;
    sub(&nMin, &box->min, &half);
    Vector3 nMax;
    add(&nMax, &box->max, &half);
    for(size_t i = 0; i < 3; i++) {
        if(nMin.data[i] > nMax.data[i]) {
            nMin.data[i] = (box->min.data[i] + box->max.data[i]) * 0.5f;
            nMax.data[i] = nMin.data[i];
        }
    }
    box->min = nMin;
    box->max = nMax;
}
